from .bitstamp import BitstampFetcher
from .cex import CexFetcher
from .kaiko import KaikoFetcher
from .coinbase import CoinbaseFetcher
from .ascendex import AscendexFetcher
from .okx import OkxFetcher
from .coingecko import CoingeckoFetcher
from .gemini import GeminiFetcher
from .thegraph import TheGraphFetcher
from .defillama import DefillamaFetcher
